
from scaleup import *

def test_create_simulation_no_values():
    actual = create_simulation_scenario('scenName', 'dsName')
    assert actual == 'scenName~dsName'


def test_create_simulation_no_values_escape_tilde():
    actual = create_simulation_scenario('scen~Name', 'dsName')
    assert actual == 'scen&TLD;Name~dsName'
    


def test_create_simulation_param_values():
    actual = create_simulation_scenario('scenName', 'dsName', 1, 2, 3, 4, 5)
    assert actual == 'scenName~dsName~1~2~3~4~5'


# no options selected, defaults to showing the UI
def test_create_script_option_string_defaults():
    actual = create_script_option_string()
    assert actual == 'SHOWUI:TRUE'


# show ui, default reader
def test_create_script_option_string_show_ui():
    actual = create_script_option_string(True)
    assert actual == 'SHOWUI:TRUE'


# hide ui, default reader
def test_create_script_option_string_hide_ui():
    actual = create_script_option_string(False)
    assert actual == ''


# show ui, alt reader
def test_create_script_option_string_altreader():
    actual = create_script_option_string(altReader = True)
    assert actual == 'SHOWUI:TRUE~ALTREADER:TRUE'


    # show ui, alt reader
def test_create_script_option_string_hide_ui_altreader():
    actual = create_script_option_string(False, True)
    assert actual == 'ALTREADER:TRUE'


def test_get_run_type():
    file_names = ['sim1.xls', 'sim1.xlsx', 'sim1.old.xlsx', 'sim1.xls.old', 'sim1.rxm', 'sim1.old.rxm']

    actual = [get_run_type(file_name) for file_name in file_names]
    expected = [1,1,1,0,2,2]

    assert expected == actual
  


"""Helper routines for COM automation of Dynochem Runtime scripting"""


import pythoncom 
from win32com.client import VARIANT

from scaleup.dynochem_automation_defs import *
from scaleup.script_utils import *

# Creates a script parameters class, adding the standard params
#   targetType of 1 is Excel 
#   targetType of 2 is Reaction Lab

def create_script_parameters(name : str , file_name : str, options : str = '', dataSheets = None):
	"""Creates a script parameters class

	:param name: Name of the automation job
	:type name: str
	:param file_name: Name of the model file
	:type file_name: str
	:param options: Options e.g. Show/Hide the UI, defaults to ''
	:type options: str, optional
	:param dataSheets: Custom Datasheets, defaults to None
	:type dataSheets: [type], optional
	:raises ValueError: If the model is not of a supported type (must be .xls? or .rxm)
	:return: COM object of script parameters
	:rtype: COM object
	"""

	target_type = get_run_type(file_name)

	if target_type == 0:
		raise ValueError('{} is not of a supported model type'.format(file_name))

	parameters = create_script_parameters_w32()
	parameters.Name = name
	parameters.TargetType = target_type
	parameters.ModelDetails = file_name
	parameters.Options = options
	if(dataSheets != None):
		set_datasheets(parameters, dataSheets)
	return parameters


def create_simulation_step(scenarios = [], writeProfiles : bool = False, returnProfiles : bool = False) -> pythoncom.VT_DISPATCH:
	"""Creates a simulation step

	:param scenarios: The scenarios to be run, If empty, all are run, defaults to []
	:type scenarios: list, optional
	:param writeProfiles: Write the profiles to an file, defaults to False
	:type writeProfiles: bool, optional
	:param returnProfiles: return the profiles as well as the endpoints, defaults to False
	:type returnProfiles: bool, optional
	:return: Script step
	:rtype: COM Object
	"""
	step = create_script_step_w32("SIMULATION")
	stepParameters = create_simulation_parameters_w32()
	
	stepParameters.Scenarios = scenarios
	
	stepParameters.RunType = 1 if len(scenarios) == 0 else 3
	stepParameters.writeProfiles = writeProfiles
	stepParameters.ReturnProfiles = returnProfiles
	step.StepParameters = stepParameters
	return step

def create_fitting_step(scenarios, parameters , fitEachScenario : bool, updateToSource : bool = False ) -> pythoncom.VT_DISPATCH:
	"""Creates a Fitting step

	:param scenarios: The scenarios to be fitted
	:type scenarios: list
	:param parameters: The parameters to be fitted
	:type parameters: list
	:param fitEachScenario: Fit to each scenario or not
	:type fitEachScenario: bool
	:param updateToSource: Write the fitted values back to the model, defaults to False
	:type updateToSource: bool, optional
	:return: Script step
	:rtype: COM Object
	"""
	step = create_script_step_w32("FITTING")
	stepParameters = create_fitting_parameters_w32()
	
	stepParameters.Scenarios = scenarios
	stepParameters.Parameters = parameters
	stepParameters.FitEachScenario = fitEachScenario
	stepParameters.UpdateToSource = updateToSource
	
	step.StepParameters = stepParameters
	return step
	
	
def create_verification_step(scenarios = [], verificationType : int = 1 ) -> pythoncom.VT_DISPATCH:
	"""Creates a Verification step

	:param scenarios: The scenarios to be run. If empty, all scenarios are run, defaults to []
	:type scenarios: list, optional
	:param verificationType: The Verification Type. 1 for full report, 2 for residuals, 3 for SSQs, defaults to 1
	:type verificationType: int, optional
	:return: Script step
	:rtype: COM Object
	"""
	step = create_script_step_w32("VERIFICATION")
	stepParameters = create_verification_parameters_w32()
	stepParameters.Scenarios = scenarios
	stepParameters.RunType = 1 if len(scenarios) == 0 else 3
	stepParameters.VerificationType = verificationType
	
	step.StepParameters = stepParameters
	return step
	
	
def create_optimization_step(scenarioName : str, factors, responses, updateToSource : bool = False) -> pythoncom.VT_DISPATCH:
	"""Creates an Optimization step

	:param scenarioName: The name of the scenario to be used
	:type scenarioName: str
	:param factors: The factors to vary
	:type factors: list
	:param responses: The Responses to be optimized
	:type responses: list
	:param updateToSource: Whether the optimized scenario and data sheet are written back to the model, defaults to False
	:type updateToSource: bool, optional
	:return: Script step
	:rtype: COM Object
	"""
	step = create_script_step_w32("OPTIMIZATION")
	stepParameters = create_optimization_parameters_w32()
	stepParameters.ScenarioName = scenarioName
	stepParameters.Factors = factors
	stepParameters.Responses = responses
	stepParameters.UpdateToSource = updateToSource
	
	step.StepParameters = stepParameters
	return step	

def set_datasheets(script, datasheets):
	datasheets_variant = VARIANT(pythoncom.VT_ARRAY | pythoncom.VT_VARIANT, datasheets)
	script.DataSheets = datasheets_variant

#=================================================================================================
# ScriptType generated from filename
	
def run_script(file_name : str, script_name : str, steps, options : str = '', dataSheets = None) -> pythoncom.VT_DISPATCH:
	"""Runs a script

	:param file_name: The file name for the model
	:type file_name: str
	:param script_name: The name of the job - displayed in the UI if visible
	:type script_name: str
	:param steps: The steps to be executed
	:type steps: list
	:param options: Option string for run (Show/Hide UI, Use Excel interop or not for Excel Models), defaults to ''
	:type options: str, optional
	:param dataSheets: Replacement Datasheets to inject data into the model, defaults to None
	:type dataSheets: list, optional
	:return: Script Result
	:rtype: COM Object
	"""

	parameters = create_script_parameters(script_name, file_name, options, dataSheets)
	parameters.Steps = steps

	runner = create_model_automate_w32()
	result = runner.RunScript(file_name,
								parameters)

	return result




from scaleup.scripts import *
from scaleup.script_utils import *
from scaleup.dynochem_automation import *


def run_simulation(file_name :str, scenarios = [], return_profiles : bool = False, show_progress = True, alt_reader = True):
    """Runs a Simulation step

    :param file_name: name of the model file
    :type file_name: str
    :param scenarios: String representations of the scenarios to be run. If empty, runs all scenarios in the model, defaults to []
    :type scenarios: list, optional
    :param return_profiles: Return the profiles as well as the endpoints, defaults to False
    :type return_profiles: bool, optional
    :param show_progress: Show the progress indicator when running, defaults to True
    :type show_progress: bool, optional
    :param alt_reader: Use the alternate reader for Excel Models (i.e. not Excel Interop), defaults to True
    :type alt_reader: bool, optional
    :return: ScriptResult object
    :rtype: ScriptResult object
    """
    options = create_script_option_string(show_progress, alt_reader)
    step = create_simulation_step(scenarios, returnProfiles=return_profiles)
    result = run_script(file_name, "Simulation Automation", [step], options)

    ret = ScriptResult([step], result)

    return ret


def run_verification(file_name : str, scenarios = [], verification_type : int = 1, show_progress = True, alt_reader = True):
    """Runs a Verification step

    :param file_name: name of the model file
    :type file_name: str
    :param scenarios: String representations of the scenarios to be run. If empty, runs all scenarios in the model, defaults to []
    :type scenarios: list, optional
    :param verification_type: Verification type. 1 for full report, 2 for residuals, 3 for SSQs, defaults to 1
    :type verification_type: int, optional
    :param show_progress: Shoe the UI when running, defaults to True
    :type show_progress: bool, optional
    :param alt_reader: Use the alternate reader for Excel Models (i.e. not Excel Interop), defaults to True
    :type alt_reader: bool, optional
    :return: ScriptResult object
    :rtype: ScriptResult object
    """
    options = create_script_option_string(show_progress, alt_reader)
    step = create_verification_step(scenarios, verification_type)
    result = run_script(file_name, "Verification Automation", [step], options)

    ret = ScriptResult([step], result)

    return ret

def run_fitting(file_name : str, scenarios, parameters, fit_to_each : bool = False, show_progress : bool = True, alt_reader : bool = True, update_to_source : bool = False):
    """Runs a Fitting step

    :param file_name: name of the model file
    :type file_name: str
    :param scenarios: the names of the scenarios to be fitted
    :type scenarios: list
    :param parameters: string representations of the parameters to be fitted
    :type parameters: list
    :param fit_to_each: Whether to fit to each scenario, defaults to False
    :type fit_to_each: bool, optional
    :param show_progress: Shoe the UI when running, defaults to True
    :type show_progress: bool, optional
    :param alt_reader: Use the alternate reader for Excel Models (i.e. not Excel Interop), defaults to True
    :type alt_reader: bool, optional
    :param update_to_source: Write back the fitted values to the original model, defaults to False
    :type update_to_source: bool, optional
    :return: ScriptResult object
    :rtype: ScriptResult
    """
    if(update_to_source):
        xl=w32c.Dispatch("Excel.Application")
        xl.Visible = True #Update to Source requires this
        wb=xl.Workbooks.Open(file_name)
        xl.WindowState = 2

    try:

        options = create_script_option_string(show_progress, alt_reader)
        step = create_fitting_step(scenarios, parameters, fit_to_each, update_to_source)
        result = run_script(file_name, "Fitting Automation", [step], options)

        ret = ScriptResult([step], result)

    finally:
        if(update_to_source):
            wb.Close(True)
            xl.Quit()

    return ret

def run_optimization(file_name : str, scenario : str, factors, responses, show_progress : bool = True, alt_reader : bool = True, update_to_source : bool = False):
    """Runs an Optimization step

    :param file_name: name of the model file
    :type file_name: str
    :param scenario: The name of the scenario to be optimized
    :type scenario: str
    :param factors: string respresentations of the factors to be varied
    :type factors: list
    :param responses: string representations of the responses to optimize
    :type responses: list
    :param show_progress: Shoe the UI when running, defaults to True
    :type show_progress: bool, optional
    :param alt_reader: Use the alternate reader for Excel Models (i.e. not Excel Interop), defaults to True
    :type alt_reader: bool, optional
    :param update_to_source: Write back the fitted values to the original model, defaults to False
    :type update_to_source: bool, optional
    :return: ScriptResult object
    :rtype: ScriptResult
    """

    if(update_to_source):
        xl=w32c.Dispatch("Excel.Application")
        xl.Visible = True #Update to Source requires this
        wb=xl.Workbooks.Open(file_name)
        xl.WindowState = 2

    try:
        options = create_script_option_string(show_progress, alt_reader)
        step = create_optimization_step(scenario, factors, responses, update_to_source)
        result = run_script(file_name, "Optimization Automation", [step], options)

        ret = ScriptResult([step], result)

    finally:
        if(update_to_source):
            wb.Close(True)
            xl.Quit()

    return ret

def run_multiple_steps(file_name : str, steps, show_progress : bool = True, alt_reader : bool = True, update_to_source : bool = False):
    """Run multiple steps in sequence

    :param file_name: name of the model file
    :type file_name: str
    :param steps: the steps to run
    :type steps: list
    :param show_progress: Shoe the UI when running, defaults to True
    :type show_progress: bool, optional
    :param alt_reader: Use the alternate reader for Excel Models (i.e. not Excel Interop), defaults to True
    :type alt_reader: bool, optional
    :param update_to_source: Write back the fitted values to the original model, defaults to False
    :type update_to_source: bool, optional
    :return: ScriptResult object
    :rtype: ScriptResult
    """
    if(update_to_source):
        xl=w32c.Dispatch("Excel.Application")
        xl.Visible = True #Update to Source requires this
        wb=xl.Workbooks.Open(file_name)
        xl.WindowState = 2

    try:
        options = create_script_option_string(show_progress, alt_reader)
        result = run_script(file_name, "Multi-step Automation", steps, options)

        ret = ScriptResult(steps, result)
        
    finally:
        if(update_to_source):
            wb.Close(True)
            xl.Quit()

    return ret



"""COM Object creation Helpers"""
import win32com.client as w32c

#==================================================================================================

def create_model_automate_w32():
	"""Creates ModelAutomate object

	:return: ModelAutomate COM object for running scripts
	:rtype: COM Object
	"""
	automate = w32c.Dispatch("ModelAutomation.ModelAutomate")
	return automate

def create_script_parameters_w32():
	"""Creates Script parameters COM Object

	:return: Script parameters COM Object
	:rtype: COM Object
	"""
	parameters = w32c.Dispatch("ModelAutomation.ScriptParameters")
	return parameters
	
def create_script_step_w32(stepType : str):
	"""Create a script step object

	:param stepType: Creates a script step
	:type stepType: str
	:return: [description]
	:rtype: [type]
	"""
	step = w32c.Dispatch("ModelAutomation.ScriptStep")
	step.StepType = stepType
	return step
	
def create_simulation_parameters_w32():
	"""Create a simulation parameters object
	
	:return: COM Object with simulation specific parameters
	:rtype: COM Object
	"""
	stepParams = w32c.Dispatch("ModelAutomation.SimulationAutomationParameters")
	return stepParams
	
def create_verification_parameters_w32():
	"""Create a verification parameters object
	
	:return: COM Object with verification specific parameters
	:rtype: COM Object
	"""
	stepParams = w32c.Dispatch("ModelAutomation.VerificationAutomationParameters")
	return stepParams	
	
def create_fitting_parameters_w32():
	"""Create a fitting parameters object
	
	:return: COM Object with fitting specific parameters
	:rtype: COM Object
	"""
	stepParams = w32c.Dispatch("ModelAutomation.FittingAutomationParameters")
	return stepParams	

def create_optimization_parameters_w32():
	"""Create an optimization parameters object
	
	:return: COM Object with optimization specific parameters
	:rtype: COM Object
	"""
	stepParams = w32c.Dispatch("ModelAutomation.OptimizationAutomationParameters")
	return stepParams	



import pandas as pd


class ScriptResult:
    """ScriptResult contains either an ErrorMessage or a list of step results
    """

    def __init__(self, steps, script_result):
        
        if ((script_result.ErrorMessage is None) or (len(script_result.ErrorMessage) == 0)):
            self.ErrorMessage = ''
            self.steps = [self.__make_step_result(step.StepType, result) for (step, result)  in zip(steps, script_result.Results) ]
        else:
            self.ErrorMessage = script_result.ErrorMessage
            self.steps = []

    __builders = { 'SIMULATION' : (lambda res : SimulationStepResult(res)),
               'VERIFICATION' : (lambda res : VerificationStepResult(res)),
               'OPTIMIZATION' : (lambda res : OptimizationStepResult(res)),
               'FITTING' : (lambda res : FittingStepResult(res)),
    }

    def __make_step_result(self, step_type : str, step_result):
        return self.__builders[step_type](step_result)





class StepResult:
    """Base class for the StepResults
    """

    def __init__(self, type : str):
        self.type = type

    def __repr__(self): 
        return self.type 
    
    def __str__(self): 
        return self.type


class SimulationStepResult(StepResult):
    """Simulation Step Result. Contains a pandas dataframe of the endpoints and (optionally) the profiles report for each scenario.
    :param StepResult: COM Step Result to be converted to Python result
    :type StepResult: COM Object
    """

    def __init__(self, result):
        StepResult.__init__(self, "SIMULATION")
        self.profiles = []
        self.__add_endpoints(result)
        self.__add_profiles(result)

    def __add_endpoints(self, result):
        sheet = result.Entries[0].Values[0]
        headings = ['{} ({})'.format(sheet[1][0],sheet[2][0])] + \
                   ['{}.{} ({})'.format(sheet[0][r],sheet[1][r],sheet[2][r]) for r in range(1,len(sheet[0]))] 
        self.endpoints =  pd.DataFrame(sheet[3:], columns=headings)       

    def __add_profiles(self, result):
        for profile in result.Entries[0].Values[1:]:
            self.profiles.append(SimulationProfile(profile))

    


class SimulationProfile:
    """Scenario profile report for Simulation Step Result
    """

    def __init__(self, entry):
        self.scenario_name = entry[0][0]
        self.solver_steps = float(entry[1][1])
        self.total_time = float(entry[2][1])
        self.matrix_size = float(entry[3][1])
        self.plotted_points = float(entry[4][1])
        self.integration_method = entry[5][1]
        self.accuracy = float(entry[6][1])
        self.run_time = float(entry[7][1])
        
        sheet = entry[9:]
        headings = ['{} ({})'.format(sheet[1][0],sheet[2][0])] + \
                    ['{}.{} ({})'.format(sheet[0][r],sheet[1][r],sheet[2][r]) for r in range(1,len(sheet[0]))]   
        self.results =  pd.DataFrame(sheet[10:], columns=headings)       
        

    def __str__(self): 
        return "Simulation Profile:\n" + \
                "Scenario Name : {}\n".format(self.scenario_name) + \
                "Solver Steps : {}\n".format(self.solver_steps) + \
                "Total Time : {}\n".format(self.total_time) + \
                "Matrix Size : {}\n".format(self.matrix_size) + \
                "Plotted Points : {}\n".format(self.plotted_points) + \
                "Integration Method : {}\n".format(self.integration_method) + \
                "Accuracy : {}\n".format(self.accuracy) + \
                "Run Time : {}\n".format(self.run_time)

class VerificationStepResult(StepResult):
    
    def __init__(self, result):
        StepResult.__init__(self, "VERIFICATION")
        self.report = result.Entries[0].Values[0]


class FittingStepResult(StepResult):
    
    def __init__(self, result):
        StepResult.__init__(self, "FITTING")
        self.report = result.Entries[0].Values[0]


class OptimizationStepResult(StepResult):
    
    def __init__(self, result):
        StepResult.__init__(self, "OPTIMIZATION")
        self.report = result.Entries[0].Values[0]




"""Various utilities for using with scripts"""

import os
from enum import Enum

__string_delim = '~'
__opt_string_delim = '\t'
delim_escape = '&TLD;'


__writers = { 
	"SIMULATION" : (lambda writer, res : __write_simulation_step_result__(writer, res)),
	"FITTING" : (lambda writer, res : __write_report_step_result__(writer, res)),
	"VERIFICATION" : (lambda writer, res : __write_report_step_result__(writer, res)),
	"OPTIMIZATION" : (lambda writer, res : __write_report_step_result__(writer, res)),
	}

class OptTarget(Enum):
	"""Enumeration for Optimization Target"""
	Target  = 1,
	Min = 2,
	Max = 3


#=============================================================================================================
# recurrently used methods


def get_file_path(file_name : str, folder_name : str = None):
	"""Returns the path of a file relative to the current folder

	:param file_name: the name of the file
	:type file_name: str
	:param folder_name: sub-folder in which the file can be found, defaults to None
	:type folder_name: str, optional
	:return: Full path to the file
	:rtype: str
	"""

	folder = os.getcwd()
	eff_folder = '' if folder_name == None else folder_name
	ret = os.path.join(folder, eff_folder, file_name)
	return ret


#Creates the simulation scenario escaping any tildes in the scenario name
def escape_name(scenName : str):
	"""Escapes a name containing tilde (~) characters, replacing each ~ with &TLD;  
	This is	used when generating tilde delimited strings to safely send names which contains tildes.

	:param scenName: name to be escaped
	:type scenName: str
	:return: Escaped version of the input string
	:rtype: str
	"""
	ret = scenName.replace(__string_delim, delim_escape)
	return ret


def create_simulation_scenario(scen_name : str, ds_name : str, values = []):
	"""Creates a simulation scenario for use with Simulation and Verification steps.

	:param scen_name: scenario name
	:type scen_name: str
	:param ds_name: datasheet name
	:type ds_name: str
	:param values: parameter values for the scenario, defaults to []
	:type values: list, optional
	:return: String representation of the scenario for use with Simulation and Verification steps.
	:rtype: str
	"""
	scen = __string_delim.join([escape_name(scen_name), ds_name])
	scen = scen if len(values) == 0 else (scen + __string_delim + __string_delim.join(str(v) for v in values))
	return scen


# creates the option string
def create_script_option_string(showInUi : bool = True, altReader : bool = False):
	"""Generates the script option string

	:param showInUi: Show the Automation UI when running, defaults to True
	:type showInUi: bool, optional
	:param altReader: Use the alternate file reader for Excel models (i.e. not Excel Interop), defaults to False
	:type altReader: bool, optional
	:return: Formatted string of options
	:rtype: str
	"""
	ret = ''
	if(showInUi):
		ret += 'SHOWUI:TRUE'
		if (altReader):
			ret += '~'
	if (altReader):
			ret += 'ALTREADER:TRUE'	
	return ret


# creates a fitting parameter
def create_fitting_parameter(parameter_name : str, unit : str, initial_value : float, max_value : float, min_value : float, fit_to_log : bool = False):
	"""Create a fitting parameter

	:param parameter_name: name of the parameter
	:type parameter_name: str
	:param unit: unit of measure for the parameter
	:type unit: str
	:param initial_value: inital value
	:type initial_value: float
	:param max_value: max value
	:type max_value: float
	:param min_value: min value
	:type min_value: float
	:param fit_to_log: Fit to the log of the value, defaults to False
	:type fit_to_log: bool, optional
	:return: Formatted parameter details for use with fitting step
	:rtype: str
	"""
	ret = __string_delim.join([parameter_name, unit, str(initial_value), str(max_value), str(min_value), str(fit_to_log)])
	return ret


# creates an optimization factor
def create_optimization_factor(name : str, unit : str, initial_value : float , min = None, max = None):
	"""Create optimization Factor

	:param name: Factor name
	:type name: str
	:param unit: factor unit
	:type unit: str
	:param initial_value: initial value
	:type initial_value: float
	:param min: min value for factor. If not set, uses half of the initial value
	:type min: float, optional
	:param max: max value for factor. If not set, uses twice the initial value
	:type max: float, optional
	:return: Formatted factor for use in optimization step
	:rtype: str
	"""
	ret = __opt_string_delim.join([name, str(initial_value), unit, str(min if min != None else initial_value / 2.0 ), str(max if max != None else initial_value * 2.0)])
	return ret


# creates an optimization response
def create_min_optimization_response(name : str, weighting : float = 1.0):
	"""Creates a 'Minimize' target for a response

	:param name: name of the response to minimize
	:type name: str
	:param weighting: Weighting for this parameter, defaults to 1.0
	:type weighting: float, optional
	:return: Formatted response for use in the Optimization step
	:rtype: str
	"""
	return create_optimization_response(name, OptTarget.Min, 0, weighting)


def create_max_optimization_response(name : str, weighting : float = 1.0):
	"""Creates a 'Maximize' target for a response

	:param name: name of the response to maximize
	:type name: str
	:param weighting: Weighting for this parameter, defaults to 1.0
	:type weighting: float, optional
	:return: Formatted response for use in the Optimization step
	:rtype: str
	"""
	return create_optimization_response(name, OptTarget.Max, 0, weighting)

def create_target_optimization_response(name : str, value : float, weighting : float = 1.0):
	"""Creates a 'Target' target for a response 

	:param name: name of the response
	:type name: str
	:param value: Target value for the response
	:type value: float
	:param weighting: Weighting for this parameter, defaults to 1.0
	:type weighting: float, optional
	:return: Formatted response for use in the Optimization step
	:rtype: str
	"""
	return create_optimization_response(name , OptTarget.Target, value , weighting)

def create_optimization_response(name : str, target : OptTarget, value : float = 0.0, weighting : float = 1.0):
	"""Create a response for the optimization step

	:param name: name of the response
	:type name: str
	:param target: Type of response. Min|Max|Target
	:type target: OptTarget
	:param value: Target Value, defaults to 0.0
	:type value: float, optional
	:param weighting: Weighting for the parameter, defaults to 1.0
	:type weighting: float, optional
	:return: Formatted response for use in the Optimization step
	:rtype: str
	"""
	ret = __opt_string_delim.join([name, str(value), str(weighting), target.name])
	return ret			



# Checks the result to see if there is an error message. if there is, it displays it,
# if not, it displays the pages of results 

def write_result(result, writer = None, separator : str = ','):
	"""
	Write the contents of a script result.
	A script result contain have an error message or an array of step results.
	If it contains an error message, then this is written
	If it contains step results, each result (each of which may contain multiple pages) is written as
	lines of values separated by the passed in separator or a comma.

	If no 'writer' method is passed in, then we print to console.

	:param result: Script Result object from RunScript()
	:type result: COM Object
	:param writer: Method which takes in a string, if not set, 'print' is used.	
	:type writer: method
	:param separator: separator used for joining values in a row. Default is ','
	:type separator: str
	"""

	if writer == None:
		writer = print

	if(result.ErrorMessage != None and len(result.ErrorMessage) > 0):
		writer(result.ErrorMessage)
	else:
		stepNumber = 1
		for stepResult in result.Results:
			writer('Step : {}\n'.format(stepNumber))
			for values in stepResult.Entries[0].Values:
				output = ''
				for row in values:
					output += separator.join(["" if v == None else str(v) for v in row]) + '\n'
				writer(output + '==\n')



def write_result_to_file(result, file_name : str, append : bool = True, separator : str = ','):
	"""
	Write the contents of a script result.
	A script result contain have an error message or an array of step results.
	If it contains an error message, then this is written
	If it contains step results, each result (each of which may contain multiple pages) is written as
	lines of values separated by the passed in separator or a comma.

	:param result: Script Result object from RunScript()
	:type result: COM Object
	:param file_name: name of the file to which we will write
	:type file_name: str
	:param append: Append the output to the file (or overwrite). Default is True (append)
	:type append: bool
	:param separator: separator used for joining values in a row. Default is ','
	:type separator: str
	"""

	try:
		file_path = get_file_path(file_name)
		mode = "a" if append else "w"
		f=open(file_path, mode)
		write_result(result, f.write, separator)
	finally:
		f.flush()
		f.close()


def __write_simulation_profile(writer, profile):
	writer('Scenario name\t{}'.format(profile.scenario_name))
	writer('Solver steps\t{}'.format(profile.solver_steps))
	writer('Total time\t{}'.format(profile.total_time))
	writer('Matrix size\t{}'.format(profile.matrix_size))
	writer('Plotted points\t{}'.format(profile.plotted_points))
	writer('Integration Method\t{}'.format(profile.integration_method))
	writer('Accuracy\t{}'.format(profile.accuracy))
	writer('Run Time\t{}'.format(profile.run_time))
	writer('\n')
	writer(profile.results)
	writer('\n')


def __write_simulation_step_result__(writer, step_result):
	writer(step_result.endpoints)
	writer('\n\n')
	for profile in step_result.profiles:
		__write_simulation_profile(writer, profile)


def __write_report_step_result__(writer, step_result, separator : str = '\t'):
	for row in step_result.report:
		output = separator.join(["" if v == None else str(v) for v in row])
		writer(output)
	writer('==\n')		



def write_script_result(result, writer = None):
	"""
	Write the contents of a script result.
	A script result contain have an error message or a list of step results.
	If it contains an error message, then this is written
	If it contains step results, each result (each of which may contain multiple pages) is written.

	If no 'writer' method is passed in, then we print to console.

	:param result: ScriptResult object from RunScript()
	:type result: ScriptResult
	:param writer: Method which takes in a string, if not set, 'print' is used.	
	:type writer: method
	"""

	if writer == None:
		writer = print

	if(result.ErrorMessage != None and len(result.ErrorMessage) > 0):
		writer(result.ErrorMessage)
	else:
		step_number = 1
		for step_result in result.steps:
			writer('Step : {} ({})\n'.format(step_number, step_result.type))
			__writers[step_result.type](writer, step_result)
			step_number += 1



def write_script_result_to_file(result, file_name : str, append : bool = True):
	"""
	Write the contents of a script result.
	A script result contain have an error message or an array of step results.
	If it contains an error message, then this is written
	If it contains step results, each result (each of which may contain multiple pages) is written 
	
	:param result: ScriptResult object from RunScript()
	:type result: ScriptResult
	:param file_name: name of the file to which we will write
	:type file_name: str
	:param append: Append the output to the file (or overwrite). Default is True (append)
	:type append: bool
	:param separator: separator used for joining values in a row. Default is ','
	:type separator: str
	"""

	try:
		file_path = get_file_path(file_name)
		mode = "a" if append else "w"
		f=open(file_path, mode)
		write_script_result(result, f.write)
	finally:
		f.flush()
		f.close()



def get_run_type(file_name : str):
	"""	Ascertains the run type of a file based on the extension.
	Excel models "*.xls(?)" are 1.
	Reaction Lab Models "*.rxm" are 2
	All others return 0

	:param file_name: name of the file
	:type file_name: str
	:return: 0|1|2, depending on the file extension
	:rtype: int
	"""
	
	file_ext = os.path.splitext(file_name)[1]
	ret = 0
	if file_ext.lower().startswith('.xls'):
		ret = 1
	if file_ext.lower().startswith('.rxm'):
		ret = 2
	return ret

